@@ -8,6 +8,8 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <assert.h>
+#include <fcntl.h>
+#include <sys/file.h>
 
 static void
 disco_trap(int sig)
@@ -172,11 +174,11 @@ usage(FILE *out, const char *name)
           " -i, --ignorecase      match items case insensitively.\n"
           " -F, --filter          filter entries for a given string before showing the menu.\n"
           " -w, --wrap            wraps cursor selection.\n"
-          " -l, --list            list items vertically with the given number of lines.\n"
+          " -l, --list            list items vertically down or up with the given number of lines(number of lines down/up). (down (default), up)\n"
           " -p, --prompt          defines the prompt text to be displayed.\n"
           " -P, --prefix          text to show before highlighted item.\n"
           " -I, --index           select item at index automatically.\n"
-          " -x, --password        hide input.\n"
+          " -x, --password        display/hide/replace input. (none (default), hide, indicator)\n"
           " -s, --no-spacing      disable the title spacing on entries.\n"
           " -C, --no-cursor       ignore cursor events.\n"
           " -T, --no-touch        ignore touch events.\n"
@@ -186,17 +188,20 @@ usage(FILE *out, const char *name)
           " --scrollbar           display scrollbar. (none (default), always, autohide)\n"
           " --counter             display a matched/total items counter. (none (default), always)\n"
           " -e, --vim-esc-exits   exit bemenu when pressing escape in normal mode for vim bindings\n"
+          " -N, --vim-normal-mode start in normal mode for vim bindings\n"
           " --accept-single       immediately return if there is only one item.\n"
           " --ifne                only display menu if there are items.\n"
+          " --single-instance     force a single menu instance.\n"
           " --fork                always fork. (bemenu-run)\n"
-          " --no-exec             do not execute command. (bemenu-run)\n\n"
+          " --no-exec             do not execute command. (bemenu-run)\n"
+          " --auto-select         when one entry is left, automatically select it\n\n"
 
           "Use BEMENU_BACKEND env variable to force backend:\n"
           " curses               ncurses based terminal backend\n"
           " wayland              wayland backend\n"
-          " x11                  x11 backend\n\n"
+          " x11                  x11 backend\n\n", out);
 
-          "If BEMENU_BACKEND is empty, one of the GUI backends is selected automatically.\n\n"
+    fputs("If BEMENU_BACKEND is empty, one of the GUI backends is selected automatically.\n\n"
 
           "Backend specific options\n"
           "   c = ncurses, w == wayland, x == x11\n"
@@ -220,6 +225,8 @@ usage(FILE *out, const char *name)
           " --tf                  defines the title foreground color. (wx)\n"
           " --fb                  defines the filter background color. (wx)\n"
           " --ff                  defines the filter foreground color. (wx)\n"
+          " --cb                  defines the cursor background color. (wx)\n"
+          " --cf                  defines the cursor foreground color. (wx)\n"
           " --nb                  defines the normal background color. (wx)\n"
           " --nf                  defines the normal foreground color. (wx)\n"
           " --hb                  defines the highlighted background color. (wx)\n"
@@ -272,13 +279,16 @@ do_getopt(struct client *client, int *argc, char **argv[])
         { "prompt",       required_argument, 0, 'p' },
         { "index",        required_argument, 0, 'I' },
         { "prefix",       required_argument, 0, 'P' },
-        { "password",     no_argument,       0, 'x' },
+        { "password",     required_argument, 0, 'x' },
         { "fixed-height", no_argument,       0, 0x090 },
         { "scrollbar",    required_argument, 0, 0x100 },
         { "counter",      required_argument, 0, 0x10a },
         { "vim-esc-exits",no_argument,       0, 'e' },
+        { "vim-normal-mode",no_argument,     0, 'N'},
         { "accept-single",no_argument,       0, 0x11a },
+        { "auto-select",  no_argument,       0, 0x11b },
         { "ifne",         no_argument,       0, 0x117 },
+        { "single-instance",no_argument,     0, 0x11c},
         { "fork",         no_argument,       0, 0x118 },
         { "no-exec",      no_argument,       0, 0x119 },
         { "bottom",       no_argument,       0, 'b' },
@@ -330,9 +340,9 @@ do_getopt(struct client *client, int *argc, char **argv[])
     for (optind = 0;;) {
         int32_t opt;
 
-        if ((opt = getopt_long(*argc, *argv, "hviwxcl:I:p:P:I:bfF:m:H:M:W:B:R:nsCTK", opts, NULL)) < 0)
+        if ((opt = getopt_long(*argc, *argv, "hviwcl:I:p:P:I:x:bfF:m:H:M:W:B:R:nsCTK", opts, NULL)) < 0)
             break;
-
+        
         switch (opt) {
             case 'h':
                 usage(stdout, *argv[0]);
@@ -340,7 +350,6 @@ do_getopt(struct client *client, int *argc, char **argv[])
             case 'v':
                 version(*argv[0]);
                 break;
-
             case 'i':
                 client->filter_mode = BM_FILTER_MODE_DMENU_CASE_INSENSITIVE;
                 break;
@@ -351,8 +360,12 @@ do_getopt(struct client *client, int *argc, char **argv[])
                 client->wrap = true;
                 break;
             case 'l':
-                client->lines = strtol(optarg, NULL, 10);
+            {
+                char *ptr;
+                client->lines = strtol(optarg, &ptr, 10);
+                client->lines_mode = (!strcmp(ptr + 1, "up") ? BM_LINES_UP : BM_LINES_DOWN);
                 break;
+            }
             case 'c':
                 client->center = true;
                 break;
@@ -377,12 +390,21 @@ do_getopt(struct client *client, int *argc, char **argv[])
             case 'e':
                 client->vim_esc_exits = true;
                 break;
+            case 'N':
+                client->vim_init_mode_normal = true;
+                break;
             case 0x11a:
                 client->accept_single = true;
                 break;
+            case 0x11b:
+                client->auto_select = true;
+                break;
             case 0x117:
                 client->ifne = true;
                 break;
+            case 0x11c:
+                client->single_instance = true;
+                break;
             case 0x118:
                 client->force_fork = true;
                 break;
@@ -390,9 +412,8 @@ do_getopt(struct client *client, int *argc, char **argv[])
                 client->no_exec = true;
                 break;
             case 'x':
-                client->password = true;
+                client->password = (!strcmp(optarg, "none") ? BM_PASSWORD_NONE : (!strcmp(optarg, "hide") ? BM_PASSWORD_HIDE : (!strcmp(optarg, "indicator") ? BM_PASSWORD_INDICATOR : BM_PASSWORD_NONE)));
                 break;
-
             case 'b':
                 client->bottom = true;
                 break;
@@ -427,10 +448,10 @@ do_getopt(struct client *client, int *argc, char **argv[])
                 client->width_factor = strtof(optarg, NULL);
                 break;
             case 'B':
-                client->border_size = strtol(optarg, NULL, 10);
+                client->border_size = strtod(optarg, NULL);
                 break;
             case 'R':
-                client->border_radius = strtol(optarg, NULL, 10);
+                client->border_radius = strtod(optarg, NULL);
                 break;
             case 0x120:
                 client->cursor_height = strtol(optarg, NULL, 10);
@@ -543,6 +564,10 @@ menu_with_options(struct client *client)
     if (!(menu = bm_menu_new(NULL)))
         return NULL;
 
+    if (client->vim_init_mode_normal) {
+        menu->vim_mode = 'n';
+    }
+
     client->fork = (client->force_fork || (bm_renderer_get_priorty(bm_menu_get_renderer(menu)) != BM_PRIO_TERMINAL));
 
     bm_menu_set_font(menu, client->font);
@@ -554,6 +579,7 @@ menu_with_options(struct client *client)
     bm_menu_set_prefix(menu, client->prefix);
     bm_menu_set_filter_mode(menu, client->filter_mode);
     bm_menu_set_lines(menu, client->lines);
+    bm_menu_set_lines_mode(menu, client->lines_mode);
     bm_menu_set_wrap(menu, client->wrap);
     bm_menu_set_monitor(menu, client->monitor);
     bm_menu_set_monitor_name(menu, client->monitor_name);
@@ -593,6 +619,20 @@ menu_with_options(struct client *client)
 enum bm_run_result
 run_menu(const struct client *client, struct bm_menu *menu, void (*item_cb)(const struct client *client, struct bm_item *item))
 {
+    if (client->single_instance) {
+        char *xdg_runtime_dir = getenv("XDG_RUNTIME_DIR");
+
+        char buffer[1024];
+        char *bemenu_lock_location = (xdg_runtime_dir == NULL ? "/tmp" : xdg_runtime_dir);
+        snprintf(buffer, sizeof(buffer), "%s/bemenu.lock", bemenu_lock_location);
+        
+        int menu_lock = open(buffer, O_CREAT | O_RDWR | O_CLOEXEC, 0666); // Create a read-write lock file(if not already exists), closes on exec. Creates in tmp due to permissions.
+        if (flock(menu_lock, LOCK_EX | LOCK_NB) == -1) {  // If we cannot set the lock/the lock is already present.
+            fprintf(stderr, "bemenu instance is already running");
+            exit(EXIT_FAILURE);
+        }
+    }
+    
     bm_menu_set_highlighted_index(menu, client->selected);
     bm_menu_grab_keyboard(menu, true);
     bm_menu_set_filter(menu, client->initial_filter);
@@ -606,7 +646,7 @@ run_menu(const struct client *client, struct bm_menu *menu, void (*item_cb)(cons
         return BM_RUN_RESULT_CANCEL;
     }
 
-    if (client->accept_single && item_count == 1) {
+    if ((client->accept_single || client->auto_select) && item_count == 1) {
         item_cb(client, *items);
         return BM_RUN_RESULT_SELECTED;
     }
@@ -619,6 +659,18 @@ run_menu(const struct client *client, struct bm_menu *menu, void (*item_cb)(cons
     struct bm_touch touch = {0};
     enum bm_run_result status = BM_RUN_RESULT_RUNNING;
     do {
+        if(client->auto_select) {
+            uint32_t item_count;
+            bm_menu_get_filtered_items(menu, &item_count);
+            if(item_count == 1) {
+                struct bm_item *highlighted = bm_menu_get_highlighted_item(menu);
+                if (highlighted) {
+                    item_cb(client, highlighted);
+                    return BM_RUN_RESULT_SELECTED;
+                }
+            }
+        }
+
         if (!bm_menu_render(menu)) {
             status = BM_RUN_RESULT_CANCEL;
             break;
