@@ -17,11 +17,12 @@ struct client {
     uint32_t cursor_width;
     uint32_t hpadding;
     uint32_t lines;
+    enum bm_lines_mode lines_mode;
     uint32_t selected;
     uint32_t monitor;
     uint32_t hmargin_size;
-    uint32_t border_size;
-    uint32_t border_radius;
+    double border_size;
+    double border_radius;
     float width_factor;
     bool bottom;
     bool center;
@@ -30,8 +31,11 @@ struct client {
     bool fixed_height; 
     bool counter;
     bool vim_esc_exits; 
+    bool vim_init_mode_normal;
     bool accept_single;
+    bool auto_select;
     bool ifne;
+    bool single_instance;
     bool no_overlap;
     bool no_spacing;
     bool no_cursor;
@@ -39,7 +43,7 @@ struct client {
     bool no_keyboard;
     bool force_fork, fork;
     bool no_exec;
-    bool password;
+    enum bm_password_mode password;
     enum bm_key_binding key_binding;
     char *monitor_name;
 };
