@@ -231,6 +231,20 @@ bm_menu_get_lines(struct bm_menu *menu)
     return menu->lines;
 }
 
+void
+bm_menu_set_lines_mode(struct bm_menu *menu, enum bm_lines_mode mode)
+{
+    assert(menu);
+    menu->lines_mode = mode;
+}
+
+enum bm_lines_mode
+bm_menu_get_lines_mode(struct bm_menu *menu)
+{
+    assert(menu);
+    return menu->lines_mode;
+}
+
 void
 bm_menu_set_wrap(struct bm_menu *menu, bool wrap)
 {
@@ -347,13 +361,13 @@ bm_menu_get_hpadding(struct bm_menu *menu)
 }
 
 void
-bm_menu_set_border_size(struct bm_menu* menu, uint32_t border_size)
+bm_menu_set_border_size(struct bm_menu* menu, double border_size)
 {
     assert(menu);
     menu->border_size = border_size;
 }
 
-uint32_t
+double
 bm_menu_get_border_size(struct bm_menu* menu)
 {
     assert(menu);
@@ -361,13 +375,13 @@ bm_menu_get_border_size(struct bm_menu* menu)
 }
 
 void
-bm_menu_set_border_radius(struct bm_menu* menu, uint32_t border_radius)
+bm_menu_set_border_radius(struct bm_menu* menu, double border_radius)
 {
     assert(menu);
     menu->border_radius = border_radius;
 }
 
-uint32_t
+double
 bm_menu_get_border_radius(struct bm_menu* menu)
 {
     assert(menu);
@@ -503,6 +517,27 @@ bm_menu_get_align(struct bm_menu *menu)
     return menu->align;
 }
 
+void
+bm_menu_set_y_offset(struct bm_menu *menu, int32_t y_offset)
+{
+    assert(menu);
+
+    if(menu->y_offset == y_offset)
+        return;
+
+    menu->y_offset = y_offset;
+
+    if (menu->renderer->api.set_y_offset)
+        menu->renderer->api.set_y_offset(menu, y_offset);
+}
+
+int32_t
+bm_menu_get_y_offset(struct bm_menu *menu)
+{
+    assert(menu);
+    return menu->y_offset;
+}
+
 void
 bm_menu_set_width(struct bm_menu *menu, uint32_t margin, float factor)
 {
@@ -616,17 +651,17 @@ bm_menu_set_spacing(struct bm_menu *menu, bool spacing)
 }
 
 void
-bm_menu_set_password(struct bm_menu *menu, bool password)
+bm_menu_set_password(struct bm_menu *menu, enum bm_password_mode mode)
 {
     assert(menu);
 
-    if (menu->password == password)
+    if (menu->password == mode)
         return;
 
-    menu->password = password;
+    menu->password = mode;
 }
 
-bool
+enum bm_password_mode
 bm_menu_get_password(struct bm_menu *menu)
 {
     assert(menu);
@@ -1041,11 +1076,11 @@ bm_menu_run_with_key(struct bm_menu *menu, enum bm_key key, uint32_t unicode)
             break;
 
         case BM_KEY_UP:
-            menu_prev(menu, count, menu->wrap);
+            (menu->lines_mode == BM_LINES_UP ? menu_next(menu, count, menu->wrap) : menu_prev(menu, count, menu->wrap));
             break;
 
         case BM_KEY_DOWN:
-            menu_next(menu, count, menu->wrap);
+            (menu->lines_mode == BM_LINES_UP ? menu_prev(menu, count, menu->wrap) : menu_next(menu, count, menu->wrap));
             break;
 
         case BM_KEY_PAGE_UP:
@@ -1117,21 +1152,61 @@ bm_menu_run_with_key(struct bm_menu *menu, enum bm_key key, uint32_t unicode)
 
         case BM_KEY_PASTE:
             {
-                FILE *clipboard;
-                clipboard = popen("wl-paste -t text/plain", "r");
-                if (!clipboard) {
-                    clipboard = popen("xclip -t text/plain -out", "r");
+                char *paster[] = {"wl-paste -t text/plain", "xclip -t text/plain -out"};
+                for (size_t paster_i = 0; paster_i < sizeof paster / sizeof paster[0]; paster_i++) {
+                    FILE *clipboard = popen(paster[paster_i], "r");
+                    if (clipboard == NULL) {
+                        fprintf(stderr, "popen() for '%s' failed: ", paster[paster_i]);
+                        perror(NULL);
+                        continue;
+                    }
+
+                    char *line = NULL;
+                    size_t len = 0;
+                    ssize_t nread;
+
+                    char *pasted_buf = NULL;
+                    size_t pasted_len = 0;
+
+                    /* Buffer all output first. If the child doesn't exit
+                     * with 0, we won't use this output and assume that it
+                     * was garbage. */
+                    while ((nread = getline(&line, &len, clipboard)) != -1) {
+                        pasted_buf = realloc(pasted_buf, pasted_len + (size_t)nread);
+                        if (pasted_buf == NULL) {
+                            fprintf(stderr, "realloc() while reading from paster '%s' failed: ", paster[paster_i]);
+                            perror(NULL);
+                            continue;
+                        }
+
+                        memmove(pasted_buf + pasted_len, line, (size_t)nread);
+                        pasted_len += (size_t)nread;
+                    }
+
+                    free(line);
+
+                    int wstatus = pclose(clipboard);
+                    if (WIFEXITED(wstatus) && WEXITSTATUS(wstatus) == 0) {
+                        for (size_t i = 0; i < pasted_len; i++) {
+                            size_t width;
+                            menu->cursor += bm_unicode_insert(&menu->filter, &menu->filter_size, menu->cursor, pasted_buf[i], &width);
+                            menu->curses_cursor += width;
+                        }
+                        free(pasted_buf);
+                        break;
+                    } else if (wstatus == -1) {
+                        fprintf(stderr, "Waiting for clipboard paste program '%s' failed: ", paster[paster_i]);
+                        perror(NULL);
+                    } else {
+                        if (WIFEXITED(wstatus)) {
+                            fprintf(stderr, "Clipboard paste program '%s' exited with %d, discarding\n", paster[paster_i], WEXITSTATUS(wstatus));
+                        } else {
+                            fprintf(stderr, "Clipboard paste program '%s' died abnormally, discarding\n", paster[paster_i]);
+                        }
+                    }
+
+                    free(pasted_buf);
                 }
-                if (!clipboard) {
-                    break;
-                }
-                int c;
-                while ((c = fgetc(clipboard)) != EOF) {
-                    size_t width;
-                    menu->cursor += bm_unicode_insert(&menu->filter, &menu->filter_size, menu->cursor, c, &width);
-                    menu->curses_cursor += width;
-                }
-                pclose(clipboard);
             }
             break;
 
