@@ -12,7 +12,7 @@ struct cairo {
     cairo_t *cr;
     cairo_surface_t *surface;
     PangoContext *pango;
-    int scale;
+    double scale;
 };
 
 struct cairo_color {
@@ -26,6 +26,10 @@ struct cairo_paint {
     int32_t baseline;
     uint32_t cursor;
     uint32_t cursor_height;
+    uint32_t cursor_width;
+    struct cairo_color cursor_fg;
+    struct cairo_color cursor_bg;
+    uint32_t hpadding;
     bool draw_cursor;
 
     struct box {
@@ -63,6 +67,8 @@ bm_cairo_create_for_surface(struct cairo *cairo, cairo_surface_t *surface)
     if (!(cairo->pango = pango_cairo_create_context(cairo->cr)))
         goto fail;
 
+    cairo_set_antialias(cairo->cr, CAIRO_ANTIALIAS_NONE);
+
     cairo->surface = surface;
     assert(cairo->scale > 0);
     cairo_surface_set_device_scale(surface, cairo->scale, cairo->scale);
@@ -154,19 +160,23 @@ bm_cairo_draw_line_str(struct cairo *cairo, struct cairo_paint *paint, struct ca
             rect.width = result.x_advance * PANGO_SCALE;
         }
 
+        uint32_t cursor_width = rect.width / PANGO_SCALE;
+        if (paint->cursor_width > 0) {
+            cursor_width = paint->cursor_width;
+        }
         uint32_t cursor_height = fmin(paint->cursor_height == 0 ? line_height : paint->cursor_height, line_height);
-        cairo_set_source_rgba(cairo->cr, paint->fg.r, paint->fg.b, paint->fg.g, paint->fg.a);
+        cairo_set_source_rgba(cairo->cr, paint->cursor_fg.r, paint->cursor_fg.b, paint->cursor_fg.g, paint->cursor_fg.a);
         cairo_rectangle(cairo->cr,
                 paint->pos.x + paint->box.lx + rect.x / PANGO_SCALE, paint->pos.y - paint->box.ty + ((line_height - cursor_height) / 2),
-                rect.width / PANGO_SCALE, cursor_height);
+                cursor_width, cursor_height);
         cairo_fill(cairo->cr);
 
         cairo_rectangle(cairo->cr,
                 paint->pos.x + paint->box.lx + rect.x / PANGO_SCALE, paint->pos.y - paint->box.ty,
-                rect.width / PANGO_SCALE, line_height);
+                cursor_width, line_height);
         cairo_clip(cairo->cr);
 
-        cairo_set_source_rgba(cairo->cr, paint->bg.r, paint->bg.b, paint->bg.g, paint->bg.a);
+        cairo_set_source_rgba(cairo->cr, paint->cursor_bg.r, paint->cursor_bg.b, paint->cursor_bg.g, paint->cursor_bg.a);
         cairo_move_to(cairo->cr, paint->box.lx + paint->pos.x, paint->pos.y - base + paint->baseline);
         pango_cairo_show_layout(cairo->cr, layout);
         cairo_reset_clip(cairo->cr);
@@ -250,7 +260,19 @@ bm_cairo_entry_message(char *entry_text, bool highlighted, uint32_t event_feedba
 }
 
 static inline void
-bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const struct bm_menu *menu, struct cairo_paint_result *out_result)
+bm_cairo_draw_rounded_path(cairo_t *cr, double x, double y, double width, double height, double radius)
+{
+    double degrees = M_PI / 180;
+    cairo_new_sub_path(cr);
+    cairo_arc(cr, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
+    cairo_arc(cr, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
+    cairo_arc(cr, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
+    cairo_arc(cr, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
+    cairo_close_path(cr);
+}
+
+static inline void
+bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct bm_menu *menu, struct cairo_paint_result *out_result)
 {
     assert(cairo && menu && out_result);
 
@@ -269,9 +291,17 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
     ascii_height = result.height;
     paint.baseline = result.baseline;
 
+    double border_size = menu->border_size;
+    width -= border_size;
     uint32_t height = fmin(fmax(menu->line_height, ascii_height), max_height);
     uint32_t vpadding = (height - ascii_height)/2;
 
+    double border_radius = menu->border_radius;
+
+    uint32_t total_item_count = menu->items.count;
+    uint32_t filtered_item_count;
+    bm_menu_get_filtered_items(menu, &filtered_item_count);
+
     cairo_set_source_rgba(cairo->cr, 0, 0, 0, 0);
     cairo_rectangle(cairo->cr, 0, 0, width, height);
 
@@ -280,45 +310,72 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
     cairo_paint(cairo->cr);
     cairo_restore(cairo->cr);
 
+    uint32_t count;
+    struct bm_item **items = bm_menu_get_filtered_items(menu, &count);
+    uint32_t lines = (menu->lines > 0 ? menu->lines : 1);
+    uint32_t page_length = 0;
+
+    if (menu->lines_mode == BM_LINES_UP && !menu->fixed_height) {
+        int32_t new_y_offset = (count < lines ? (lines - count) * height : 0);
+        bm_menu_set_y_offset(menu, new_y_offset);
+    }
+
     memset(&result, 0, sizeof(result));
     uint32_t title_x = 0;
+
     if (menu->title) {
-        bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_FG, &paint.fg);
-        bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_BG, &paint.bg);
-        paint.pos = (struct pos){ result.x_advance, vpadding };
-        paint.box = (struct box){ 4, 8, vpadding, -vpadding, 0, height };
+        /* Essentially hide the title text if we are drawing lines "up". */
+        enum bm_color title_fg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_TITLE_FG);
+        enum bm_color title_bg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_TITLE_BG);
+        bm_cairo_color_from_menu_color(menu, title_fg, &paint.fg);
+        bm_cairo_color_from_menu_color(menu, title_bg, &paint.bg);
+        paint.pos = (struct pos){ border_size + 4, vpadding + border_size };
+        paint.box = (struct box){ 4, 16, vpadding, -vpadding, 0, height };
         bm_cairo_draw_line(cairo, &paint, &result, "%s", menu->title);
         title_x = result.x_advance;
     }
 
-    bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_FG, &paint.fg);
-    bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_BG, &paint.bg);
+    enum bm_color filter_fg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_FILTER_FG);
+    enum bm_color filter_bg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_FILTER_BG);
+    enum bm_color cursor_fg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_CURSOR_FG);
+    enum bm_color cursor_bg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_CURSOR_BG);
+    bm_cairo_color_from_menu_color(menu, filter_fg, &paint.fg);
+    bm_cairo_color_from_menu_color(menu, filter_bg, &paint.bg);
+    bm_cairo_color_from_menu_color(menu, cursor_fg, &paint.cursor_fg);
+    bm_cairo_color_from_menu_color(menu, cursor_bg, &paint.cursor_bg);
     paint.draw_cursor = true;
     paint.cursor = menu->cursor;
     paint.cursor_height = menu->cursor_height;
-    paint.pos = (struct pos){ (menu->title ? 2 : 0) + result.x_advance, vpadding };
+    paint.cursor_width = menu->cursor_width;
+    paint.pos = (struct pos){ (menu->title ? 2 : 0) + result.x_advance + border_size, vpadding + border_size };
     paint.box = (struct box){ (menu->title ? 2 : 4), 0, vpadding, -vpadding, width - paint.pos.x, height };
 
-    const char *filter_text = (menu->filter ? menu->filter : "");
-    if (menu->password) {
-        bm_cairo_draw_line_str(cairo, &paint, &result, "");
-    } else {
-        bm_cairo_draw_line(cairo, &paint, &result, "%s", filter_text);
+    if (menu->lines == 0 || menu->lines_mode == BM_LINES_DOWN) {
+        const char *filter_text = (menu->filter ? menu->filter : "");
+        if (menu->password == BM_PASSWORD_HIDE) {
+            bm_cairo_draw_line_str(cairo, &paint, &result, "");
+        } else if (menu->password == BM_PASSWORD_INDICATOR) {
+            char asterisk_print[1024] = "";
+        
+            for (int i = 0; i < (int)(strlen(filter_text)); ++i) asterisk_print[i] = '*';
+        
+            bm_cairo_draw_line(cairo, &paint, &result, "%s", asterisk_print);
+        } else if (menu->password == BM_PASSWORD_NONE) {
+            bm_cairo_draw_line(cairo, &paint, &result, "%s", filter_text);
+        }
     }
 
     paint.draw_cursor = false;
-    const uint32_t titleh = result.height;
+    uint32_t titleh = (result.height > 0 ? result.height : height);
     out_result->height = titleh;
 
-    uint32_t count;
-    struct bm_item **items = bm_menu_get_filtered_items(menu, &count);
-    uint32_t lines = (menu->lines > 0 ? menu->lines : 1);
+    uint32_t posy = (menu->lines_mode == BM_LINES_UP ? 0 : titleh);
+    uint32_t spacing_x = menu->spacing ? title_x : 0, spacing_y = 0; // 0 == variable width spacing
 
     if (menu->lines > 0) {
         /* vertical mode */
 
         const bool scrollbar = (menu->scrollbar > BM_SCROLLBAR_NONE && (menu->scrollbar != BM_SCROLLBAR_AUTOHIDE || count > lines) ? true : false);
-        uint32_t spacing_x = menu->spacing ? title_x : 0, spacing_y = 0; // 0 == variable width spacing
         if (lines > max_height / titleh) {
             /* there is more lines than screen can fit, enter fixed spacing mode */
             lines = max_height / titleh - 1;
@@ -338,10 +395,24 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
             spacing_x += (spacing_x < scrollbar_w ? scrollbar_w : 0);
         }
 
-        uint32_t posy = titleh;
         const uint32_t page = (menu->index / lines) * lines;
-        for (uint32_t l = 0, i = page; l < lines && i < count && posy < max_height; ++i, ++l) {
-            bool highlighted = (items[i] == bm_menu_get_highlighted_item(menu));
+        
+        for (uint32_t l = 0, i = page; l < lines && posy < max_height; ++i, ++l) {
+            if (!menu->fixed_height && i >= count) {
+                continue;
+            }
+
+
+            uint32_t is_fixed_up = (menu->fixed_height && menu->lines_mode == BM_LINES_UP);
+            uint32_t required_empty = (count > lines ? 0 : lines - count);
+            int32_t last_item_index = (count < lines ? count - 1 + page : lines - 1 + page);
+            int32_t display_item_index = (menu->lines_mode == BM_LINES_DOWN ? i : menu->fixed_height ? last_item_index - i + page + required_empty  : last_item_index - i + page);
+            display_item_index = (display_item_index < 0 ? 0 : display_item_index);
+
+            bool highlighted = false;
+            if ((i < count && !is_fixed_up) || (is_fixed_up && display_item_index <= last_item_index)) {
+                highlighted = (items[display_item_index] == bm_menu_get_highlighted_item(menu));
+            }
 
             if (highlighted) {
                 if (menu->event_feedback) {
@@ -351,21 +422,28 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
                     bm_cairo_color_from_menu_color(menu, BM_COLOR_HIGHLIGHTED_FG, &paint.fg);
                     bm_cairo_color_from_menu_color(menu, BM_COLOR_HIGHLIGHTED_BG, &paint.bg);
                 }
-            } else if (bm_menu_item_is_selected(menu, items[i])) {
+            } else if (i < count && bm_menu_item_is_selected(menu, items[i])) {
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_SELECTED_FG, &paint.fg);
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_SELECTED_BG, &paint.bg);
+            } else if (i < count && i % 2 == 1) {
+                bm_cairo_color_from_menu_color(menu, BM_COLOR_ALTERNATE_FG, &paint.fg);
+                bm_cairo_color_from_menu_color(menu, BM_COLOR_ALTERNATE_BG, &paint.bg);
             } else {
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_FG, &paint.fg);
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
             }
+            
+            char *line_str = "";
+            if ((i < count && !is_fixed_up) || (is_fixed_up && display_item_index <= last_item_index)) {
+                line_str = bm_cairo_entry_message(items[display_item_index]->text, highlighted, menu->event_feedback, i,  count);
+            }
 
-            char *line_str = bm_cairo_entry_message(items[i]->text, highlighted, menu->event_feedback, i,  count);
             if (menu->prefix && highlighted) {
-                paint.pos = (struct pos){ spacing_x, posy+vpadding };
+                paint.pos = (struct pos){ spacing_x + border_size, posy + vpadding + border_size };
                 paint.box = (struct box){ 4, 0, vpadding, -vpadding, width - paint.pos.x, height };
                 bm_cairo_draw_line(cairo, &paint, &result, "%s %s", menu->prefix, line_str);
             } else {
-                paint.pos = (struct pos){ spacing_x, posy+vpadding };
+                paint.pos = (struct pos){ spacing_x + border_size, posy+vpadding + border_size };
                 paint.box = (struct box){ 4 + prefix_x, 0, vpadding, -vpadding, width - paint.pos.x, height };
                 bm_cairo_draw_line(cairo, &paint, &result, "%s", line_str);
             }
@@ -373,13 +451,14 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
             posy += (spacing_y ? spacing_y : result.height);
             out_result->height = posy;
             out_result->displayed++;
+            page_length += 1;
         }
 
         if (spacing_x) {
             bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
             const uint32_t sheight = out_result->height - titleh;
             cairo_set_source_rgba(cairo->cr, paint.bg.r, paint.bg.b, paint.bg.g, paint.bg.a);
-            cairo_rectangle(cairo->cr, scrollbar_w, titleh, spacing_x - scrollbar_w, sheight);
+            cairo_rectangle(cairo->cr, scrollbar_w + border_size, titleh + border_size, spacing_x - scrollbar_w, sheight);
             cairo_fill(cairo->cr);
         }
 
@@ -389,14 +468,15 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
 
             const uint32_t sheight = out_result->height - titleh;
             cairo_set_source_rgba(cairo->cr, paint.bg.r, paint.bg.b, paint.bg.g, paint.bg.a);
-            cairo_rectangle(cairo->cr, 0, titleh, scrollbar_w, sheight);
+            cairo_rectangle(cairo->cr, border_size, titleh + border_size, scrollbar_w, sheight);
             cairo_fill(cairo->cr);
 
             const float percent = fmin(((float)page / (count - lines)), 1.0f);
-            const uint32_t size = fmax(sheight * ((float)lines / count), 2.0f);
+            const float fraction = fmin((float)lines / count, 1.0f);
+            const uint32_t size = fmax(sheight * fraction, 2.0f);
             const uint32_t posy = percent * (sheight - size);
             cairo_set_source_rgba(cairo->cr, paint.fg.r, paint.fg.b, paint.fg.g, paint.fg.a);
-            cairo_rectangle(cairo->cr, 0, titleh + posy, scrollbar_w, size);
+            cairo_rectangle(cairo->cr, border_size, titleh + posy + border_size, scrollbar_w, size);
             cairo_fill(cairo->cr);
         }
     } else {
@@ -405,13 +485,13 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
         uint32_t cl = fmin(title_x + result.x_advance, width / 4);
 
         if (count > 0) {
-            paint.pos = (struct pos){ cl, vpadding };
+            paint.pos = (struct pos){ cl, vpadding + border_size };
             paint.box = (struct box){ 1, 2, vpadding, -vpadding, 0, height };
             bm_cairo_draw_line(cairo, &paint, &result, (count > 0 && (menu->wrap || menu->index > 0) ? "<" : " "));
             cl += result.x_advance + 1;
         }
 
-        for (uint32_t i = menu->index; i < count && cl < (width/cairo->scale); ++i) {
+        for (uint32_t i = menu->index; i < count && cl < (width/cairo->scale); ++i) { 
             bool highlighted = (items[i] == bm_menu_get_highlighted_item(menu));
 
             if (highlighted) {
@@ -420,29 +500,100 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, const s
             } else if (bm_menu_item_is_selected(menu, items[i])) {
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_SELECTED_FG, &paint.fg);
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_SELECTED_BG, &paint.bg);
+            } else if (i % 2 == 1) {
+                bm_cairo_color_from_menu_color(menu, BM_COLOR_ALTERNATE_FG, &paint.fg);
+                bm_cairo_color_from_menu_color(menu, BM_COLOR_ALTERNATE_BG, &paint.bg);
             } else {
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_FG, &paint.fg);
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
             }
 
-            paint.pos = (struct pos){ cl, vpadding };
-            paint.box = (struct box){ 2, 4, vpadding, -vpadding, 0, height };
+            uint32_t hpadding = (menu->hpadding == 0 ? 2 : menu->hpadding);
+            paint.pos = (struct pos){ cl + (hpadding/2), vpadding + border_size };
+            paint.box = (struct box){ hpadding/2, 1.5 * hpadding, vpadding, -vpadding, 0, height };
             bm_cairo_draw_line(cairo, &paint, &result, "%s", (items[i]->text ? items[i]->text : ""));
-            cl += result.x_advance + 2;
+            cl += result.x_advance + (0.5 * hpadding);
             out_result->displayed += (cl < width);
             out_result->height = fmax(out_result->height, result.height);
         }
-
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_FG, &paint.fg);
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_BG, &paint.bg);
         if (menu->wrap || menu->index + 1 < count) {
-            bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_FG, &paint.fg);
-            bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_BG, &paint.bg);
             bm_pango_get_text_extents(cairo, &paint, &result, ">");
-            paint.pos = (struct pos){ width/cairo->scale - result.x_advance - 2, vpadding };
+            paint.pos = (struct pos){ width/cairo->scale - result.x_advance - 2, vpadding + border_size };
             paint.box = (struct box){ 1, 2, vpadding, -vpadding, 0, height };
             bm_cairo_draw_line(cairo, &paint, &result, ">");
         }
+        
     }
 
+    if (menu->lines_mode == BM_LINES_UP) {
+        if (menu->title) {
+            bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_FG, &paint.fg);
+            bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_BG, &paint.bg);
+        
+            paint.pos = (struct pos){ border_size + 4, posy + vpadding + border_size };
+            paint.box = (struct box){ 4, 16, vpadding, -vpadding, 0, height };
+            bm_cairo_draw_line(cairo, &paint, &result, "%s", menu->title);
+        }
+
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_FG, &paint.fg);
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_FILTER_BG, &paint.bg);
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_CURSOR_FG, &paint.cursor_fg);
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_CURSOR_BG, &paint.cursor_bg);
+        paint.draw_cursor = true;
+        paint.cursor = menu->cursor;
+        paint.cursor_height = menu->cursor_height;
+        paint.cursor_width = menu->cursor_width;
+        result.x_advance = (menu->title ? result.x_advance : 0);
+        paint.pos = (struct pos){ (menu->title ? 2 : 0) + result.x_advance + border_size, posy + border_size };
+        paint.box = (struct box){ (menu->title ? 2 : 4), 0, vpadding, -vpadding, width - paint.pos.x, height };
+
+        const char *filter_text = (menu->filter ? menu->filter : "");
+        if (menu->password == BM_PASSWORD_HIDE) {
+            bm_cairo_draw_line_str(cairo, &paint, &result, "");
+        } else if (menu->password == BM_PASSWORD_INDICATOR) {
+            char asterisk_print[1024] = "";
+        
+            for (int i = 0; i < (int)(strlen(filter_text)); ++i) asterisk_print[i] = '*';
+        
+            bm_cairo_draw_line(cairo, &paint, &result, "%s", asterisk_print);
+        } else if (menu->password == BM_PASSWORD_NONE) {
+            bm_cairo_draw_line(cairo, &paint, &result, "%s", filter_text);
+        }
+
+        paint.draw_cursor = false;
+
+        posy += (spacing_y ? spacing_y : result.height);
+        out_result->height = posy;
+        out_result->displayed++;
+    }
+
+    if (menu->counter) {
+        char counter[128];
+        snprintf(counter, sizeof(counter), "[%u/%u]", filtered_item_count, total_item_count);
+        bm_pango_get_text_extents(cairo, &paint, &result, "%s", counter);
+        
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_FG, &paint.fg);
+        bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
+        paint.pos = (struct pos){ width/cairo->scale - result.x_advance - 10, vpadding + border_size };
+        paint.box = (struct box){ 1, 2, vpadding, -vpadding, 0, height };
+        bm_cairo_draw_line(cairo, &paint, &result, "%s", counter);
+    }
+
+    // Draw borders
+    bm_cairo_color_from_menu_color(menu, BM_COLOR_BORDER, &paint.fg);
+    cairo_set_source_rgba(cairo->cr, paint.fg.r, paint.fg.b, paint.fg.g, paint.fg.a);
+    if (!border_radius) {
+        cairo_rectangle(cairo->cr, 0, 0, (width + border_size)/cairo->scale, (height * (page_length + 1)) + (2 * border_size));
+    } else {
+        bm_cairo_draw_rounded_path(cairo->cr, 0, 0, (width + border_size)/cairo->scale, (height * (page_length + 1)) + (2 * border_size), border_radius);
+    }
+    cairo_set_line_width(cairo->cr, 2 * menu->border_size);
+    
+    cairo_stroke(cairo->cr);
+
+    out_result->height += 2 * border_size;
     out_result->height *= cairo->scale;
 }
 
