@@ -14,7 +14,7 @@
 static int efd;
 
 static void
-render_windows_if_pending(const struct bm_menu *menu, struct wayland *wayland) {
+render_windows_if_pending(struct bm_menu *menu, struct wayland *wayland) {
     struct window *window;
     wl_list_for_each(window, &wayland->windows, link) {
         if (window->render_pending)
@@ -23,14 +23,12 @@ render_windows_if_pending(const struct bm_menu *menu, struct wayland *wayland) {
     wl_display_flush(wayland->display);
 }
 
-static void
+static bool
 wait_for_events(struct wayland *wayland) {
     wl_display_dispatch_pending(wayland->display);
 
-    if (wl_display_flush(wayland->display) < 0 && errno != EAGAIN) {
-        wayland->input.sym = XKB_KEY_Escape;
-        return;
-    }
+    if (wl_display_flush(wayland->display) < 0 && errno != EAGAIN)
+        return false;
 
     struct epoll_event ep[16];
     uint32_t num = epoll_wait(efd, ep, 16, -1);
@@ -38,11 +36,13 @@ wait_for_events(struct wayland *wayland) {
         if (ep[i].data.ptr == &wayland->fds.display) {
             if (ep[i].events & EPOLLERR || ep[i].events & EPOLLHUP ||
                ((ep[i].events & EPOLLIN) && wl_display_dispatch(wayland->display) < 0))
-                wayland->input.sym = XKB_KEY_Escape;
+                return false;
         } else if (ep[i].data.ptr == &wayland->fds.repeat) {
             bm_wl_repeat(wayland);
         }
     }
+
+    return true;
 }
 
 static void
@@ -54,7 +54,8 @@ schedule_windows_render_if_dirty(struct bm_menu *menu, struct wayland *wayland)
             // be(re)created. We need to do the render ASAP (not schedule it) because otherwise,
             // since we lack a window, we may not receive further events and will get deadlocked
             render_windows_if_pending(menu, wayland);
-        } else if (menu->dirty) {
+        }
+        if (menu->dirty) {
             bm_wl_window_schedule_render(window);
         }
     }
@@ -62,14 +63,17 @@ schedule_windows_render_if_dirty(struct bm_menu *menu, struct wayland *wayland)
     menu->dirty = false;
 }
 
-static void
+static bool
 render(struct bm_menu *menu)
 {
     struct wayland *wayland = menu->renderer->internal;
 
     schedule_windows_render_if_dirty(menu, wayland);
-    wait_for_events(wayland);
+    if (!wait_for_events(wayland))
+        return false;
     render_windows_if_pending(menu, wayland);
+
+    return true;
 }
 
 static enum bm_key
@@ -86,9 +90,6 @@ poll_key(const struct bm_menu *menu, unsigned int *unicode)
     uint32_t mods = wayland->input.modifiers;
     *unicode = xkb_state_key_get_utf32(wayland->input.xkb.state, wayland->input.code);
 
-    if (!*unicode && wayland->input.code == 23 && (mods & MOD_SHIFT))
-        return BM_KEY_SHIFT_TAB;
-
     wayland->input.sym = XKB_KEY_NoSymbol;
     wayland->input.code = 0;
 
@@ -136,6 +137,7 @@ poll_key(const struct bm_menu *menu, unsigned int *unicode)
         case XKB_KEY_Delete:
             return (mods & MOD_SHIFT ? BM_KEY_LINE_DELETE_LEFT : BM_KEY_DELETE);
 
+        case XKB_KEY_ISO_Left_Tab: //shift tab gives this sym
         case XKB_KEY_Tab:
             return (mods & MOD_SHIFT ? BM_KEY_SHIFT_TAB : BM_KEY_TAB);
 
@@ -167,6 +169,9 @@ poll_key(const struct bm_menu *menu, unsigned int *unicode)
         case XKB_KEY_l:
             return (mods & MOD_CTRL ? BM_KEY_LEFT : (mods & MOD_ALT ? BM_KEY_DOWN : BM_KEY_UNICODE));
 
+        case XKB_KEY_b:
+            return (mods & MOD_CTRL ? BM_KEY_LEFT : BM_KEY_UNICODE);
+
         case XKB_KEY_f:
             return (mods & MOD_CTRL ? BM_KEY_RIGHT : BM_KEY_UNICODE);
 
@@ -245,7 +250,7 @@ poll_pointer(const struct bm_menu *menu)
     struct pointer_event *event = &input->pointer_event;
     assert(wayland && event);
 
-    struct bm_pointer bm_pointer;
+    struct bm_pointer bm_pointer = {0};
 
     bm_pointer.event_mask = event->event_mask;
     bm_pointer.pos_x = wl_fixed_to_int(event->surface_x);
@@ -267,10 +272,10 @@ poll_pointer(const struct bm_menu *menu)
     }
 
     if (event->state & WL_POINTER_BUTTON_STATE_PRESSED) {
-        bm_pointer.state |= POINTER_STATE_PRESSED;
+        bm_pointer.state = POINTER_STATE_PRESSED;
     }
     if (event->state & WL_POINTER_BUTTON_STATE_RELEASED) {
-        bm_pointer.state |= POINTER_STATE_RELEASED;
+        bm_pointer.state = POINTER_STATE_RELEASED;
     }
 
     memset(event, 0, sizeof(*event));
@@ -393,6 +398,18 @@ set_align(const struct bm_menu *menu, enum bm_align align)
     }
 }
 
+static void
+set_y_offset(const struct bm_menu *menu, int32_t y_offset)
+{
+    struct wayland *wayland = menu->renderer->internal;
+    assert(wayland);
+
+    struct window *window;
+    wl_list_for_each(window, &wayland->windows, link) {
+        bm_wl_window_set_y_offset(window, wayland->display, y_offset);
+    }
+}
+
 static void
 grab_keyboard(const struct bm_menu *menu, bool grab)
 {
@@ -417,6 +434,50 @@ set_overlap(const struct bm_menu *menu, bool overlap)
     }
 }
 
+void
+wl_surface_enter(void *data, struct wl_surface *wl_surface,
+                 struct wl_output *wl_output)
+{
+    (void)wl_surface;
+    struct window *window = data;
+    struct wayland *wayland = window->wayland;
+
+    struct output *output;
+    wl_list_for_each(output, &wayland->outputs, link) {
+        if (output->output == wl_output) {
+            struct surf_output *surf_output = calloc(1, sizeof(struct surf_output));
+            surf_output->output = output;
+            wl_list_insert(&window->surf_outputs, &surf_output->link);
+            break;
+        }
+    }
+
+    window_update_output(window);
+}
+
+void
+wl_surface_leave(void *data, struct wl_surface *wl_surface,
+                 struct wl_output *wl_output)
+{
+    (void)wl_surface;
+    struct window *window = data;
+
+    struct surf_output *surf_output;
+    wl_list_for_each(surf_output, &window->surf_outputs, link) {
+        if (surf_output->output->output == wl_output) {
+            wl_list_remove(&surf_output->link);
+            break;
+        }
+    }
+
+    window_update_output(window);
+}
+
+static const struct wl_surface_listener surface_listener = {
+    .enter = wl_surface_enter,
+    .leave = wl_surface_leave,
+};
+
 static void
 destroy_windows(struct wayland *wayland)
 {
@@ -427,58 +488,74 @@ destroy_windows(struct wayland *wayland)
     wl_list_init(&wayland->windows);
 }
 
-static void
+void
+window_update_output(struct window *window)
+{
+    int32_t max_scale = 1;
+    uint32_t min_max_height = 0;
+
+    struct surf_output *surf_output;
+    wl_list_for_each(surf_output, &window->surf_outputs, link) {
+        if (surf_output->output->scale > max_scale) {
+            max_scale = surf_output->output->scale;
+        }
+        if (min_max_height == 0 || surf_output->output->height < min_max_height) {
+            min_max_height = surf_output->output->height;
+        };
+    }
+
+    if (min_max_height != window->max_height) {
+        window->max_height = min_max_height;
+    }
+
+    if (max_scale != window->scale) {
+        window->scale = max_scale;
+    }
+}
+
+void
 recreate_windows(const struct bm_menu *menu, struct wayland *wayland)
 {
     destroy_windows(wayland);
 
-    int32_t monitors = 0;
-    struct output *output;
-    wl_list_for_each(output, &wayland->outputs, link)
-        monitors++;
+    struct window *window = calloc(1, sizeof(struct window));
+    wl_list_init(&window->surf_outputs);
+    window->wayland = wayland;
+    window->align = menu->align;
+    window->hmargin_size = menu->hmargin_size;
+    window->width_factor = menu->width_factor;
 
-    int32_t monitor = 0;
-    wl_list_for_each(output, &wayland->outputs, link) {
+    // TODO: this should not be necessary, but Sway 1.8.1 does not trigger event
+    // surface.enter before we actually need to render the first frame.
+    window->scale = 1;
+    window->max_height = 640;
 
-        if (!menu->monitor_name) {
-            if (menu->monitor > -1) {
-                 if (menu->monitor < monitors && monitor != menu->monitor) {
-                    ++monitor;
-                    continue;
-                }
-            }
-        } else if (strcmp(menu->monitor_name, output->name)) {
-            continue;
-        }
+    struct wl_surface *surface = NULL;
+    if (!(surface = wl_compositor_create_surface(wayland->compositor)))
+        goto fail;
 
-        struct wl_surface *surface;
-        if (!(surface = wl_compositor_create_surface(wayland->compositor)))
-            goto fail;
+    wl_surface_add_listener(surface, &surface_listener, window);
 
-        wl_surface_set_buffer_scale(surface, output->scale);
+    struct output *output = NULL;
+    if (wayland->selected_output) {
+        fprintf(stderr, "selected output\n");
+        output = wayland->selected_output;
+    };
 
-        struct window *window = calloc(1, sizeof(struct window));
-        window->align = menu->align;
-        window->hmargin_size = menu->hmargin_size;
-        window->width_factor = menu->width_factor;
+    struct wl_output *wl_output = NULL;
+    if (output)
+        wl_output = output->output;
 
-        const char *scale = getenv("BEMENU_SCALE");
-        if (scale) {
-            window->scale = fmax(strtof(scale, NULL), 1.0f);
-        } else {
-            window->scale = output->scale;
-        }
-
-        if (!bm_wl_window_create(window, wayland->display, wayland->shm, (menu->monitor == -1) ? NULL : output->output, wayland->layer_shell, surface))
-            free(window);
-
-        window->notify.render = bm_cairo_paint;
-        window->max_height = output->height;
-        window->render_pending = true;
-        wl_list_insert(&wayland->windows, &window->link);
-        if (menu->monitor != -2) break;
+    if (!bm_wl_window_create(window, wayland->display, wayland->shm,
+                             wl_output, wayland->layer_shell, surface)) {
+        free(window);
+        goto fail;
     }
 
+    window->notify.render = bm_cairo_paint;
+    window->render_pending = true;
+    wl_list_insert(&wayland->windows, &window->link);
+
     set_overlap(menu, menu->overlap);
     grab_keyboard(menu, menu->grabbed);
     return;
@@ -504,7 +581,19 @@ set_monitor_name(const struct bm_menu *menu, char *monitor_name)
     (void)monitor_name;
     struct wayland *wayland = menu->renderer->internal;
     assert(wayland);
-    recreate_windows(menu, wayland);
+
+    if (!monitor_name) {
+        return;
+    }
+
+    struct output *output;
+    wl_list_for_each(output, &wayland->outputs, link) {
+        if (0 == strcmp(monitor_name, output->name)) {
+            wayland->selected_output = output;
+            recreate_windows(menu, wayland);
+            return;
+        }
+    }
 }
 
 static void
@@ -593,6 +682,7 @@ register_renderer(struct render_api *api)
     api->release_touch = release_touch;
     api->render = render;
     api->set_align = set_align;
+    api->set_y_offset = set_y_offset;
     api->set_width = set_width;
     api->grab_keyboard = grab_keyboard;
     api->set_overlap = set_overlap;
